@@ -38,6 +38,55 @@ else:
         return char
     bytes_from_ints = bytes
 
+if sys.version_info < (2,6):
+    # Add getcode into 2.5
+    class _addinfourl(urllib.addinfourl):
+        def __init__(self, fp, headers, url, code=None):
+            urllib.addinfourl.__init__(self, fp, headers, url)
+            self.code = code
+        def getcode(self):
+            return self.code
+
+    class FancyURLopener(urllib.FancyURLopener):
+        def http_error_default(self, url, fp, errcode, errmsg, headers):
+            return _addinfourl(fp, headers, "http:" + url, errcode)
+else:
+    FancyURLopener = urllib.FancyURLopener
+
+class NotAuthenticated(Exception):
+    CONNECTION_REFUSED = 1
+    DIRECT_VERIFICATION_FAILED = 2
+    CANCELLED = 3
+    UNSUPPORTED_VERSION = 4
+    UNEXPECTED_MODE = 5
+    CLAIMED_ID_MISSING = 6
+    DISCOVERY_FAILED = 7
+    INCONSISTENT_IDS = 8
+    REPLAY_ATTACK = 9
+    MISSING_NONCE = 10
+    msgs = {
+        CONNECTION_REFUSED : 'OP refuses connection with status %d',
+        DIRECT_VERIFICATION_FAILED : 'OP doesn\'t assert that the signature of the verification request is valid',
+        CANCELLED : 'OP did not authenticate user (cancelled)',
+        UNSUPPORTED_VERSION : 'Unsupported OpenID version',
+        UNEXPECTED_MODE : 'Unexpected mode %s',
+        CLAIMED_ID_MISSING : 'Cannot determine claimed ID',
+        DISCOVERY_FAILED : 'Claimed ID %s cannot be rediscovered',
+        INCONSISTENT_IDS : 'Discovered and asserted endpoints differ',
+        REPLAY_ATTACK : 'Replay attack detected',
+        MISSING_NONCE : 'Nonce missing in OpenID 2 response',
+        }
+
+    def __init__(self, errno, *args):
+        msg = self.msgs[errno]
+        if args:
+            msg %= args
+        self.errno = errno
+        Exception.__init__(self, msg, errno, *args)
+
+    def __str__(self):
+        return self.args[0]
+
 def normalize_uri(uri):
     """Normalize an uri according to OpenID section 7.2. Return a pair
     type,value, where type can be either 'xri' or 'uri'."""
@@ -345,7 +394,7 @@ def btwoc(l):
     return res[3+b2i(res[3]):3:-1]
 
 def unbtwoc(B):
-    return cPickle.loads(b('\x80\x02\x8a')+bytes_from_ints([len(B)])+B[::-1]+(b'.'))
+    return cPickle.loads(b('\x80\x02\x8a')+bytes_from_ints([len(B)])+B[::-1]+b('.'))
 
 # Appendix B; DH default prime
 dh_prime = """
@@ -386,8 +435,8 @@ def associate(services, url):
         if data['openid.session_type'] == "no-encryption":
             data['openid.session_type'] = ''
         del data['openid.ns']
-    res = urllib.urlopen(url, b(urllib.urlencode(data)))
-    if res.getcode() != 200:
+    res = FancyURLopener().open(url, b(urllib.urlencode(data)))
+    if hasattr(res, 'getcode') and res.getcode() != 200:
         raise ValueError, "OpenID provider refuses connection with status %d" % res.getcode()
     data = parse_response(res.read())
     if 'error' in data:
@@ -479,10 +528,15 @@ def request_authentication(services, url, assoc_handle, return_to,
         if sreg_opt:
             data['openid.sreg.optional'] =  sreg11['openid.sreg11.optional'] =','.join(sreg_opt)
     if is_compat_1x(services):
+        # OpenID 1.1 does not communicate claimed_ids. Put them into the return URL
+        return_to += '&' if '?' in return_to else '?'
+        return_to += '&openid1=' + urllib.quote(claimed)
+        data['openid.return_to'] = return_to
         del data['openid.ns']
         del data['openid.claimed_id']
         del data['openid.realm']
-        data['openid.trust_root'] = return_to
+        trust_root = urlparse.urlparse(return_to)[:3] + (None,None,None)
+        data['openid.trust_root'] = urlparse.urlunparse(trust_root)
     ax_req, ax_opt = ax
     if "http://openid.net/srv/ax/1.0" in services and (ax_req or ax_opt):
         data.update({
@@ -505,8 +559,20 @@ def request_authentication(services, url, assoc_handle, return_to,
     else:
         return url+"?"+urllib.urlencode(data)
 
-class NotAuthenticated(Exception):
-    pass
+# 11.4.2 Verifying Directly with the OpenID Provider
+def _verify_signature_directly(op_endpoint, response):
+    '''Request that the OP verify the signature via Direct Verification'''
+
+    request = [('openid.mode', 'check_authentication')]
+    # Exact copies of all fields from the authentication response, except for
+    # "openid.mode"
+    request.extend((k, v) for k, (v,) in response.items() if 'openid.mode' != k)
+    res = FancyURLopener().open(op_endpoint, urllib.urlencode(request))
+    if hasattr(res, 'getcode') and 200 != res.getcode():
+        raise NotAuthenticated(NotAuthenticated.CONNECTION_REFUSED, res.getcode())
+    response = parse_response(res.read())
+    if 'true' != response['is_valid']:
+        raise NotAuthenticated(NotAuthenticated.DIRECT_VERIFICATION_FAILED)
 
 def _prepare_response(response):
     if isinstance(response, str):
@@ -537,7 +603,7 @@ def authenticate(session, response):
     if session['assoc_handle'] != response['openid.assoc_handle'][0]:
         raise ValueError('incorrect session')
     if response['openid.mode'][0] == 'cancel':
-        raise NotAuthenticated('provider did not authenticate user (cancelled)')
+        raise NotAuthenticated(NotAuthenticated.CANCELLED)
     if response['openid.mode'][0] != 'id_res':
         raise ValueError('invalid openid.mode')
     if  'openid.identity' not in response:
@@ -577,6 +643,73 @@ def authenticate(session, response):
 
     return signed
 
+# td.total_seconds only works in 2.7
+def _total_seconds(td):
+    return td.days*24*3600 + td.seconds
+
+def verify(response, discovery_cache, find_association, nonce_seen):
+    response = _prepare_response(response)
+    if 'openid.ns' in response:
+        ns = response['openid.ns'][0]
+        if ns != 'http://specs.openid.net/auth/2.0':
+            raise NotAuthenticated(NotAuthenticate.UNSUPPORTED_VERSION)
+    else:
+        ns = None
+    mode = response['openid.mode'][0]
+    if mode == 'cancel':
+        raise NotAuthenticated(NotAuthenticated.CANCELLED)
+    if mode != 'id_res':
+        raise NotAuthenticated(NotAuthenticated.UNEXPECTED_MODE, mode)
+    # Establish claimed ID
+    if 'openid.claimed_id' in response:
+        claimed_id = response['openid.claimed_id'][0]
+        # 11.2. Drop Fragment from claimed_id
+        fragment = claimed_id.find('#')
+        if fragment != -1:
+            claimed_id = claimed_id[:fragment]
+    elif 'openid1' in response:
+        claimed_id = response['openid1'][0]
+    else:
+        raise NotAuthenticated(NotAuthenticated.CLAIMED_ID_MISSING)
+    discovered = discovery_cache(claimed_id)
+    if not discovered:
+        discovered = discover(claimed_id)
+        if not discovered:
+            raise NotAuthenticated(NotAuthenticated.DISCOVERY_FAILED, claimed_id)
+    services, op_endpoint, op_local = discovered
+    # For a provider-allocated claimed_id, there will be no op_local ID,
+    # and there is no point checking it.
+    if op_local and op_local != response['openid.identity'][0]:
+        raise NotAuthenticated('Discovered and asserted local identifiers differ')
+    # For OpenID 1.1, op_endpoint may not be included in the response
+    if ('openid.op_endpoint' in response and
+        op_endpoint != response['openid.op_endpoint'][0]):
+        raise NotAuthenticated(NotAuthenticated.INCONSISTENT_IDS)
+    # XXX verify protocol version, verify claimed_id wrt. original request,
+    # verify return_to URL
+    
+    # verify the signature
+    assoc_handle = response['openid.assoc_handle'][0]
+    session = find_association(assoc_handle)
+    if session:
+        signed = authenticate(session, response)
+    else:
+        _verify_signature_directly(op_endpoint, response)
+        signed = response['openid.signed'][0].split(',')
+
+    # Check the nonce. OpenID 1.1 doesn't have them
+    if 'openid.response_nonce' in response:
+        nonce = response['openid.response_nonce'][0]
+        timestamp = parse_nonce(nonce)
+        if _total_seconds(datetime.datetime.utcnow() - timestamp) > 300:
+            # allow for at most 300s transmission time and time shift
+            raise NotAuthenticated(NotAuthenticated.REPLAY_ATTACK)
+        if nonce_seen(nonce):
+            raise NotAuthenticated(NotAuthenticated.REPLAY_ATTACK)
+    elif ns:
+        raise NotAuthenticated(NotAuthenticated.MISSING_NONCE)
+    return signed, claimed_id
+
 def parse_nonce(nonce):
     '''Extract a datetime.datetime stamp from the nonce'''
     stamp = nonce.split('Z', 1)[0]
